@@ -266,6 +266,36 @@ incbuflen(buflen)
 		return MaxMessageSize;
 }
 
+#if _FFR_TESTS
+/* just for testing/debug output */
+static const char *
+makeprint(c)
+	char c;
+{
+	static char prt[6];
+
+	prt[1] = '\0';
+	prt[2] = '\0';
+	if (isprint((unsigned char)c))
+		prt[0] = c;
+	else if ('\n' == c)
+	{
+		prt[0] = 'L';
+		prt[1] = 'F';
+	}
+	else if ('\r' == c)
+	{
+		prt[0] = 'C';
+		prt[1] = 'R';
+	}
+	else
+		snprintf(prt, sizeof(prt), "%o", c);
+	return prt;
+}
+#else /* _FFR_TESTS */
+# define makeprint(c)	"X"
+#endif /* _FFR_TESTS */
+
 /*
 **  COLLECT -- read & parse message header & make temp file.
 **
@@ -346,16 +376,23 @@ collect(fp, smtpmode, hdrp, e, rsetsize)
 #if _FFR_REJECT_NUL_BYTE
 	bool hasNUL;		/* has at least one NUL input byte */
 #endif
-	bool bare_lf = false;
-	bool bare_cr = false;
+	int bare_lf, bare_cr;
 
 #define SMTPMODE	(smtpmode >= SMTPMODE_LAX)
 #define SMTPMODE_STRICT	((smtpmode & SMTPMODE_CRLF) != 0)
-#define SMTPMODE_BARE_LF_421	((smtpmode & SMTPMODE_LF_421) != 0)
-#define SMTPMODE_BARE_CR_421	((smtpmode & SMTPMODE_CR_421) != 0)
+#define BARE_LF_421	((smtpmode & SMTPMODE_LF_421) != 0)
+#define BARE_CR_421	((smtpmode & SMTPMODE_CR_421) != 0)
+#define BARE_LF_SP	((smtpmode & SMTPMODE_LF_SP) != 0)
+#define BARE_CR_SP	((smtpmode & SMTPMODE_CR_SP) != 0)
+
+/* for bare_{lf,cr} */
+#define BARE_IN_HDR	0x01
+#define BARE_IN_BDY	0x02
+#define BARE_WHERE	((MS_BODY == mstate) ? BARE_IN_BDY : BARE_IN_HDR)
 
 	df = NULL;
 	ignrdot = SMTPMODE ? false : IgnrDot;
+	bare_lf = bare_cr = 0;
 
 	/* timeout for I/O functions is in milliseconds */
 	dbto = SMTPMODE ? ((int) TimeOuts.to_datablock * 1000)
@@ -472,27 +509,22 @@ collect(fp, smtpmode, hdrp, e, rsetsize)
 					HasEightBits |= bitset(0x80, c);
 			}
 			if (tTd(30, 94))
-				sm_dprintf("istate=%s, c=%c (0x%x)\n",
-					ISTATE, (char) c, c);
-			if ('\n' == c && !bare_lf && SMTPMODE &&
+				sm_dprintf("istate=%s, c=%s (0x%x)\n",
+					ISTATE, makeprint((char) c), c);
+			if ('\n' == c && SMTPMODE &&
 			    !(IS_CR == istate || IS_DOTCR == istate))
 			{
-				bare_lf = true;
-				if (SMTPMODE_BARE_LF_421)
+				bare_lf |= BARE_WHERE;
+				if (BARE_LF_421)
 				{
 					inputerr = true;
 					goto readabort;
 				}
-			}
-			if (!bare_cr && SMTPMODE &&
-			    (IS_CR == istate || IS_DOTCR == istate)
-			    && '\n' != c)
-			{
-				bare_cr = true;
-				if (SMTPMODE_BARE_CR_421)
+				if (BARE_LF_SP)
 				{
-					inputerr = true;
-					goto readabort;
+					if (TTD(30, 64))
+						sm_dprintf("LF: c=%s %#x\n", makeprint((char) c), c);
+					c = ' ';
 				}
 			}
 			switch (istate)
@@ -552,12 +584,30 @@ collect(fp, smtpmode, hdrp, e, rsetsize)
 
 			  case IS_CR:
 				if (c == '\n')
+				{
+					if (TTD(30, 64))
+						sm_dprintf("state=CR, c=%s %#x -> BOL\n", makeprint((char) c), c);
 					istate = IS_BOL;
+				}
 				else
 				{
+					if (TTD(30, 64))
+						sm_dprintf("state=CR, c=%s %#x -> NORM\n", makeprint((char) c), c);
+					if (SMTPMODE)
+					{
+						bare_cr |= BARE_WHERE;
+						if (BARE_CR_421)
+						{
+							inputerr = true;
+							goto readabort;
+						}
+					}
 					(void) sm_io_ungetc(fp, SM_TIME_DEFAULT,
 							    c);
-					c = '\r';
+					if (BARE_CR_SP)
+						c = ' ';
+					else
+						c = '\r';
 					istate = IS_NORM;
 				}
 				goto bufferchar;
@@ -593,7 +643,8 @@ bufferchar:
 				if (!bitset(EF_TOOBIG, e->e_flags))
 					(void) sm_io_putc(df, SM_TIME_DEFAULT,
 							  c);
-
+				if (TTD(30, 64))
+					sm_dprintf("state=%s, put=%s %#x\n", ISTATE, makeprint((char) c), c);
 				/* FALLTHROUGH */
 
 			  case MS_DISCARD:
@@ -721,7 +772,7 @@ nextstate:
 			/* guaranteed by isheader(buf) */
 			SM_ASSERT(*(bp - 1) != '\n' || bp > buf + 1);
 
-			/* trim off trailing CRLF or NL */
+			/* trim off trailing CRLF or LF */
 			if (*--bp != '\n' || *--bp != '\r')
 				bp++;
 			*bp = '\0';
@@ -874,34 +925,33 @@ readdone:
   readabort:
 	if (inputerr && (OpMode == MD_SMTP || OpMode == MD_DAEMON))
 	{
-		char *host;
 		char *problem;
 		ADDRESS *q;
 
-		host = RealHostName;
-		if (host == NULL)
-			host = "localhost";
-
 		if (sm_io_eof(fp))
 			problem = "unexpected close";
 		else if (sm_io_error(fp))
 			problem = "I/O error";
-		else if (bare_lf)
+		else if (0 != bare_lf)
 			problem = "bare LF";
-		else if (bare_cr)
+		else if (0 != bare_cr)
 			problem = "bare CR";
 		else
 			problem = "read timeout";
 
-#define CONN_ERR_TXT	"collect: %s on connection from %s, from=%s%s%s"
+#define LOG_CLT ((NULL == RealHostName) ? RealHostName: "localhost")
+#define CONN_ERR_TXT	"collect: %s on connection from %s, from=%s%s%s%s%s"
 #define CONN_ERR_CODE	"421 4.4.1 "
-#define CONN_ERR_BARE bare_lf ? BARE_LF_MSG : (bare_cr ? BARE_CR_MSG : "")
-#define CONN_ERR_ARGS problem, host, shortenstring(e->e_from.q_paddr, MAXSHORTSTR), (bare_lf || bare_cr) ? ", info=" : "", CONN_ERR_BARE
-		if (LogLevel > 0 && (sm_io_eof(fp) || bare_lf || bare_cr))
+#define CONN_LOG_FROM	shortenstring(e->e_from.q_paddr, MAXSHORTSTR)
+#define CONN_ERR_BARE (0 != bare_lf) ? BARE_LF_MSG : ((0 != bare_cr) ? BARE_CR_MSG : "")
+#define CONN_ERR_WHERE (BARE_IN_HDR==(bare_lf|bare_cr) ? "header" : \
+	(BARE_IN_BDY==(bare_lf|bare_cr) ? "body" : "header and body"))
+#define CONN_ERR_ARGS problem, LOG_CLT, CONN_LOG_FROM, (0 != (bare_lf | bare_cr)) ? ", info=" : "", CONN_ERR_BARE, (0 != (bare_lf | bare_cr)) ? ", where=" : "", (0 != (bare_lf | bare_cr)) ? CONN_ERR_WHERE : ""
+		if (LogLevel > 0 && (sm_io_eof(fp) || (0 != (bare_lf | bare_cr))))
 			sm_syslog(LOG_NOTICE, e->e_id,
 				CONN_ERR_TXT, CONN_ERR_ARGS);
-		if (bare_lf || bare_cr)
-			usrerr(CONN_ERR_CODE "%s", CONN_ERR_BARE);
+		if (0 != (bare_lf | bare_cr))
+			usrerr("421 4.5.0 %s", CONN_ERR_BARE);
 		else if (sm_io_eof(fp))
 			usrerr(CONN_ERR_CODE CONN_ERR_TXT, CONN_ERR_ARGS);
 		else
@@ -935,10 +985,21 @@ readdone:
 		logsender(e, e->e_msgid);
 		e->e_flags &= ~EF_LOGSENDER;
 	}
-	if (bare_lf && LogLevel > 8)
-		sm_syslog(LOG_NOTICE, e->e_id, BARE_LF_MSG);
-	if (bare_cr && LogLevel > 8)
-		sm_syslog(LOG_NOTICE, e->e_id, BARE_CR_MSG);
+
+#define LOG_BARE_XY(bare_xy, bare_xy_sp, bare_xy_msg)	\
+	do	\
+	{	\
+		if ((0 != bare_xy) && LogLevel > 8)	\
+			sm_syslog(LOG_NOTICE, e->e_id, \
+				"collect: connection from %s, from=%s, info=%s, where=%s%s" \
+				, LOG_CLT, CONN_LOG_FROM, bare_xy_msg	\
+				, CONN_ERR_WHERE	\
+				, bare_xy_sp ? ", stat=replaced" : ""	\
+				);	\
+	} while (0)
+
+	LOG_BARE_XY(bare_lf, BARE_LF_SP, BARE_LF_MSG);
+	LOG_BARE_XY(bare_cr, BARE_CR_SP, BARE_CR_MSG);
 
 	/* check for message too large */
 	if (bitset(EF_TOOBIG, e->e_flags))
